"""This file contains all the Qualys API URLs used by qualyspy."""

about = "/msp/about.php/"
host_list_vm_detection = "/api/3.0/fo/asset/host/vm/detection/"
host_list = "/api/4.0/fo/asset/host/"
knowledgebase = "/api/3.0/fo/knowledge_base/vuln/"
ignore_vuln = "/api/2.0/fo/ignore_vuln/index.php/"
launch_vm_scan = "/api/2.0/fo/scan/"
map_report_list = "/msp/map_report_list.php"
download_saved_map_report = "/msp/map_report.php"

gateway_auth = "/auth/"

asset_details = "/rest/2.0/get/am/asset"
all_asset_details = "/rest/2.0/search/am/asset"

create_tag = "/qps/rest/2.0/create/am/tag"
update_tag = "/qps/rest/2.0/update/am/tag"
search_tags = "/qps/rest/2.0/search/am/tag"
count_tags = "/qps/rest/2.0/count/am/tag"
delete_tag = "/qps/rest/2.0/delete/am/tag"
get_asset_info = "/qps/rest/2.0/get/am/asset"
update_asset = "/qps/rest/2.0/update/am/asset"
search_assets = "/qps/rest/2.0/search/am/asset"

add_bulk_external_sites = "/certview/v1/externalSites/bulkAdd"



from pydantic_xml import BaseXmlModel, attr, element, wrapped


class Link(BaseXmlModel):
    value: str = attr(name="value")
    link: str | None = None


class Discovery(BaseXmlModel):
    method: str = attr(name="method")
    discovery: str | None = None


class Port(BaseXmlModel):
    value: str = attr(name="value")
    port: str | None = None


class Ip(BaseXmlModel):
    port: list[Port] | None = element(tag="PORT", default=None)
    discovery: list[Discovery] | None = element(tag="DISCOVERY", default=None)
    link: list[Link] | None = element(tag="LINK", default=None)
    value: str = attr(name="value")
    name: str | None = attr(name="name", default=None)
    type: str | None = attr(name="type", default=None)
    os: str | None = attr(name="os", default=None)
    account: str | None = attr(name="account", default=None)
    netbios: str | None = attr(name="netbios", default=None)
    network: str | None = attr(name="network", default=None)
    network_id: str | None = attr(name="network_id", default=None)


class OptionProfileTitle(BaseXmlModel):
    option_profile_default: str | None = attr("option_profile_default", default=None)
    option_profile_title: str


class OptionProfile(BaseXmlModel):
    option_profile_title: OptionProfileTitle = element(tag="OPTION_PROFILE_TITLE")


class Range(BaseXmlModel):
    start: str = element(tag="START")
    end: str = element(tag="END")


class Netblock(BaseXmlModel):
    range: list[Range] = element(tag="RANGE")


class UserEnteredDomains(BaseXmlModel):
    domain: list[str] = element(tag="DOMAIN")
    netblock: list[Netblock] | None = element(tag="NETBLOCK", default=None)


class AssetGroup(BaseXmlModel):
    asset_group_title: str = element(tag="ASSET_GROUP_TITLE")


class Key(BaseXmlModel):
    value: str = attr(name="value")
    key: str


class Header(BaseXmlModel):
    key: list[Key] = element(tag="KEY")
    asset_groups: list[AssetGroup] | None = wrapped(
        "ASSET_GROUPS", element(tag="ASSET_GROUP", default=None)
    )
    user_entered_domains: UserEnteredDomains | None = element(
        tag="USER_ENTERED_DOMAINS"
    )
    option_profile: OptionProfile | None = element(tag="OPTION_PROFILE")


class Error(BaseXmlModel):
    number: int | None = element(tag="NUMBER", default=None)
    value: str | None


class Map(BaseXmlModel, tag="MAP"):
    header: Header | None = element(tag="HEADER", default=None)
    ip: list[Ip] | None = element(tag="IP", default=None)
    error: Error | None = element(tag="ERROR", default=None)
    value: str | None = attr(name="value", default=None)



import configparser
import requests
import os

CONFIG_FILE = os.path.expanduser("~/qualysapi.conf")
config = configparser.ConfigParser()
config.read(CONFIG_FILE)
API_ROOT = config['AUTHENTICATION']['api_root']

headers = {"X-Requested-With": "qualysapi python package"}
data = {"action": "logout"}

with open("debug/cookies.txt", "r") as f:
    for cookie in f:
        cookies = {"QualysSession": cookie.strip("\n")}
        r = requests.post(API_ROOT + "/api/2.0/fo/session/", headers=headers,
                          data=data, cookies=cookies)
        print(r.text)



from pydantic_xml import BaseXmlModel, element, attr

from typing import Literal

tag_rule_types = Literal[
    "STATIC",
    "GROOVY",
    "OS_REGEX",
    "NETWORK_RANGE",
    "NETWORK_RANGE_ENHANCED" "NAME_CONTAINS",
    "INSTALLED_SOFTWARE",
    "OPEN_PORTS",
    "VULN_EXISTS",
    "ASSET_SEARCH",
    "CLOUD_ASSET",
    "BUSINESS_INFORMATION",
    "GLOBAL_ASSET_VIEW",
    "TAG_SET",
]

tag_provider_types = Literal["EC2", "AZURE", "GCP", "IBM", "OCI", "Alibaba"]


class TagSimple(BaseXmlModel, tag="TagSimple"):
    id: int | None = element(tag="id", default=None)
    name: str | None = element(tag="name", default=None)


class CollectionType(BaseXmlModel):
    tag_simple: list[TagSimple] | None = element(tag="TagSimple", default=None)


class Collection(BaseXmlModel):
    set: CollectionType | None = element(tag="set", default=None)
    add: CollectionType | None = element(tag="add", default=None)
    remove: CollectionType | None = element(tag="remove", default=None)
    update: CollectionType | None = element(tag="update", default=None)


class Tag(BaseXmlModel):
    name: str | None = element(tag="name", default=None)
    rule_type: tag_rule_types | None = element(tag="ruleType", default=None)
    rule_text: str | None = element(tag="ruleText", default=None)
    criticality_score: int | None = element(tag="criticalityScore", default=None)
    color: str | None = element(tag="color", default=None)
    children: Collection | None = element(tag="children", default=None)


class Data(BaseXmlModel):
    tag: Tag = element(tag="Tag")


class Criteria(BaseXmlModel):
    field: str = attr(name="field")
    operator: str = attr(name="operator")
    value: str


class Filters(BaseXmlModel):
    criteria: list[Criteria] = element(tag="Criteria")


class ServiceRequest(BaseXmlModel, tag="ServiceRequest"):
    data: Data | None = element(tag="data", default=None)
    filters: Filters | None = element(tag="filters", default=None)


def create_add_tag_request(
    name: str,
    rule_type: tag_rule_types | None = None,
    rule_text: str | None = None,
    criticality_score: int | None = None,
    color: str | None = None,
    children: list[str] = [],
) -> ServiceRequest:
    if children:
        children_to_add = Collection(
            set=CollectionType(tag_simple=[TagSimple(name=child) for child in children])
        )
    else:
        children_to_add = None
    return ServiceRequest(
        data=Data(
            tag=Tag(
                name=name,
                rule_type=rule_type,
                rule_text=rule_text,
                criticality_score=criticality_score,
                color=color,
                children=children_to_add,
            )
        )
    )


def create_update_tag_request(
    name: str | None = None,
    rule_type: tag_rule_types | None = None,
    rule_text: str | None = None,
    criticality_score: int | None = None,
    color: str | None = None,
    add_children: list[str] = [],
    remove_children: list[int] = [],
) -> ServiceRequest:
    if add_children or remove_children:
        children = Collection(
            set=CollectionType(
                tag_simple=[TagSimple(name=child) for child in add_children]
            ),
            remove=CollectionType(
                tag_simple=[TagSimple(id=child) for child in remove_children]
            ),
        )
    else:
        children = None

    return ServiceRequest(
        data=Data(
            tag=Tag(
                name=name,
                rule_type=rule_type,
                rule_text=rule_text,
                criticality_score=criticality_score,
                color=color,
                children=children,
            )
        )
    )


def create_search_tags_request(
    id: int | None,
    name: str | None,
    parent: int | None,
    rule_type: tag_rule_types | None,
    provider: tag_provider_types | None,
    color: str | None,
) -> ServiceRequest:
    criteria_list = []
    for field, value in {
        "id": id,
        "name": name,
        "parent": parent,
        "ruleType": rule_type,
        "provider": provider,
        "color": color,
    }.items():
        if value:
            criteria_list.append(Criteria(field=field, operator="EQUALS", value=str(value)))
    return ServiceRequest(filters=Filters(criteria=criteria_list))



from pydantic_xml import BaseXmlModel, element, wrapped, attr
import datetime
import ipaddress


class AssetGroup(BaseXmlModel):
    id: int = element(tag="ID")
    title: str = element(tag="TITLE")


class User(BaseXmlModel):
    user_login: str = element(tag="USER_LOGIN")
    first_name: str = element(tag="FIRST_NAME")
    last_name: str = element(tag="LAST_NAME")


class UserDef(BaseXmlModel):
    label_1: str | None = element(tag="LABEL_1", default=None)
    label_2: str | None = element(tag="LABEL_2", default=None)
    label_3: str | None = element(tag="LABEL_3", default=None)
    value_1: str | None = element(tag="VALUE_1", default=None)
    value_2: str | None = element(tag="VALUE_2", default=None)
    value_3: str | None = element(tag="VALUE_3", default=None)


class Glossary(BaseXmlModel):
    user_def: UserDef | None = element(tag="USER_DEF", default=None)
    user_list: list[User] | None = wrapped(
        "USER_LIST", element(tag="USER", default=None)
    )
    asset_group_list: list[AssetGroup] | None = wrapped(
        "ASSET_GROUP_LIST", element(tag="ASSET_GROUP", default=None)
    )


class ResponseWarning(BaseXmlModel):
    code: str = element(tag="CODE")
    text: str = element(tag="TEXT")
    url: str = element(tag="URL")


class IdSet(BaseXmlModel):
    id: list[int] | None = element(tag="ID", default=None)
    id_range: list[int] | None = element(tag="ID_RANGE", default=None)


class CloudTag(BaseXmlModel):
    name: str = element(tag="NAME")
    value: str = element(tag="VALUE")
    last_success_date: datetime.datetime = element(tag="LAST_SUCCESS_DATE")


class Attribute(BaseXmlModel):
    name: str = element(tag="NAME")
    last_status: str = element(tag="LAST_STATUS")
    value: str = element(tag="VALUE")
    last_success_date: datetime.datetime | None = element(
        tag="LAST_SUCCESS_DATE", default=None
    )
    last_error_date: datetime.datetime | None = element(
        tag="LAST_ERROR_DATE", default=None
    )
    last_error: str | None = element(tag="LAST_ERROR", default=None)


class Metadata(BaseXmlModel):
    ec2: list[Attribute] = wrapped("EC2", element(tag="ATTRIBUTE"))
    google: list[Attribute] = wrapped("GOOGLE", element(tag="ATTRIBUTE"))
    azure: list[Attribute] = wrapped("AZURE", element(tag="ATTRIBUTE"))


class Tag(BaseXmlModel):
    tag_id: int = element(tag="TAG_ID")
    name: str = element(tag="NAME")


class DnsData(BaseXmlModel):
    hostname: str | None = element(tag="HOSTNAME", default=None)
    domain: str | None = element(tag="DOMAIN", default=None)
    fqdn: str | None = element(tag="FQDN", default=None)


class VulnCount(BaseXmlModel):
    qds_severity: str = attr(name="qds_severity")
    count: int | None


class TruriskScoreFactors(BaseXmlModel):
    trurisk_score_formula: str | None = element(
        tag="TRURISK_SCORE_FORMULA", default=None
    )
    vuln_count: list[VulnCount] | None = element(tag="VULN_COUNT", default=None)


class Host(BaseXmlModel):
    id: int = element(tag="ID")
    asset_id: int | None = element(tag="ASSET_ID", default=None)
    ip: ipaddress.IPv4Address = element(tag="IP", default=None)
    ipv6: ipaddress.IPv6Address | None = element(tag="IPV6", default=None)
    trurisk_score: int | None = element(tag="TRURISK_SCORE", default=None)
    asset_criticality_score: int | None = element(
        tag="ASSET_CRITICALITY_SCORE", default=None
    )
    trurisk_score_factors: TruriskScoreFactors | None = element(
        tag="TRURISK_SCORE_FACTORS", default=None
    )
    tracking_method: str | None = element(tag="TRACKING_METHOD", default=None)
    network_id: int | None = element(tag="NETWORK_ID", default=None)
    dns: str | None = element(tag="DNS", default=None)
    dns_data: DnsData | None = element(tag="DNS_DATA", default=None)
    cloud_provider: str | None = element(tag="CLOUD_PROVIDER", default=None)
    cloud_service: str | None = element(tag="CLOUD_SERVICE", default=None)
    cloud_resource_id: str | None = element(tag="CLOUD_RESOURCE_ID", default=None)
    ec2_instance_id: str | None = element(tag="EC2_INSTANCE_ID", default=None)
    netbios: str | None = element(tag="NETBIOS", default=None)
    os: str | None = element(tag="OS", default=None)
    qg_hostid: str | None = element(tag="QG_HOSTID", default=None)
    last_boot: datetime.datetime | None = element(tag="LAST_BOOT", default=None)
    serial_number: str | None = element(tag="SERIAL_NUMBER", default=None)
    hardware_uuid: str | None = element(tag="HARDWARE_UUID", default=None)
    first_found_date: datetime.datetime | None = element(
        tag="FIRST_FOUND_DATE", default=None
    )
    last_activity: datetime.datetime | None = element(tag="LAST_ACTIVITY", default=None)
    agent_status: str | None = element(tag="AGENT_STATUS", default=None)
    cloud_agent_running_on: str | None = element(
        tag="CLOUD_AGENT_RUNNING_ON", default=None
    )
    tags: list[Tag] = wrapped("TAGS", element(tag="TAG", default_factory=list))
    metadata: Metadata | None = element(tag="METADATA", alias="metadata_", default=None)
    cloud_provider_tags: list[CloudTag] | None = wrapped(
        "CLOUD_PROVIDER_TAGS", element(tag="CLOUD_TAG", default=None)
    )
    last_vuln_scan_datetime: datetime.datetime | None = element(
        tag="LAST_VULN_SCAN_DATETIME", default=None
    )
    last_vm_scanned_date: datetime.datetime | None = element(
        tag="LAST_VM_SCANNED_DATE", default=None
    )
    last_vm_scanned_duration: int | None = element(
        tag="LAST_VM_SCANNED_DURATION", default=None
    )
    last_vm_auth_scanned_date: datetime.datetime | None = element(
        tag="LAST_VM_AUTH_SCANNED_DATE", default=None
    )
    last_vm_auth_scanned_duration: int | None = element(
        tag="LAST_VM_AUTH_SCANNED_DURATION", default=None
    )
    last_compliance_scan_datetime: datetime.datetime | None = element(
        tag="LAST_COMPLIANCE_SCAN_DATETIME", default=None
    )
    last_scap_scan_datetime: datetime.datetime | None = element(
        tag="LAST_SCAP_SCAN_DATETIME", default=None
    )
    owner: str | None = element(tag="OWNER", default=None)
    comments: str | None = element(tag="COMMENTS", default=None)
    user_def: UserDef | None = element(tag="USER_DEF", default=None)
    asset_group_ids: str | None = element(tag="ASSET_GROUP_IDS", default=None)


class Response(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    host_list: list[Host] | None = wrapped(
        "HOST_LIST", element(tag="HOST", default=None)
    )
    id_set: IdSet | None = element(tag="ID_SET", default=None)
    warning: ResponseWarning | None = element(tag="WARNING", default=None)
    glossary: Glossary | None = element(tag="GLOSSARY", default=None)


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    request_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] = element(tag="PARAM_LIST")
    post_data: str = element(tag="POST_DATA")


class HostListOutput(BaseXmlModel, tag="HOST_LIST_OUTPUT"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE")



# mypy: ignore-errors
# type: ignore

import inspect
import ipaddress
import os
import sys
import unittest

import sqlalchemy as sa

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from qualyspy import gav  # noqa: E402
from qualyspy.models.gav import asset_details_orm  # noqa: E402


class TestOutputModels(unittest.TestCase):
    def test_asset_details(self):
        api = gav.GavAPI()
        asset = api.asset_details(asset_id=14355608)

        self.assertEqual(asset.address, ipaddress.ip_address("172.16.76.84"))

    def test_all_asset_details(self):
        api = gav.GavAPI()
        assets, _, _ = api.all_asset_details(
            asset_filter=[{"field": "asset.assetID", "operator": "EQUALS", "value": "14355608"},
                    {"field": "inventory.source", "operator": "EQUALS", "value": "Cloud Agent"}],
            filter_operation="AND"
        )
        asset = assets[0]

        self.assertEqual(asset.address, ipaddress.ip_address("172.16.76.84"))


class TestORM(unittest.TestCase):
    def test_sql_all_asset_details(self):
        api = gav.AllAssetDetailsORM()
        api.drop()
        api.init_db()
        api.load()
        stmt = sa.select(asset_details_orm.AssetItem).where(
            asset_details_orm.AssetItem.asset_id == 14355608
        )
        result = api.query(stmt)
        asset = result[0][0]

        self.assertEqual(asset.address, ipaddress.ip_address("172.16.76.84"))



"""
Qualys VMDR API Module

Typical usage example:
vmdr_orm = vmdr.HostListDetectionORM()
vmdr_orm.load()
session = orm.Session(vmdr_orm.engine):
host_list = session.query(HostORM).all()
stmt = session.query(HostORM)
hosts = vmdr_orm.query(stmt)[0]
host = hosts.host[0]
session.close()
"""

import datetime
import ipaddress
import re
from typing import Any, Literal

import sqlalchemy.orm as orm

from . import URLS, qutils
from .base import QualysAPIBase, QualysORMMixin
from .models.vmdr import (
    host_list_orm,
    host_list_output,
    host_list_vm_detection_orm,
    host_list_vm_detection_output,
    knowledgebase_orm,
    knowledgebase_output,
    map_report,
    map_report_list,
    simple_return,
)


class VmdrAPI(QualysAPIBase):
    """Qualys VMDR API Class.  Contains methods for interacting with the VMDR API."""

    def host_list(
        self,
        *,
        show_asset_id: bool | None = None,
        details: Literal["Basic", "Basic/AGs", "All", "All/AGs", "None"] | None = None,
        os_pattern: str | None = None,
        truncation_limit: int | None = None,
        ips: (
            list[str | ipaddress.IPv4Address | ipaddress.IPv6Address]
            | str
            | ipaddress.IPv4Address
            | ipaddress.IPv6Address
            | None
        ) = None,
        ag_ids: int | list[int] | None = None,
        ag_titles: str | list[str] | None = None,
        ids: int | list[int] | None = None,
        id_min: int | None = None,
        id_max: int | None = None,
        network_ids: int | list[int] | None = None,
        compliance_enabled: bool | None = None,
        no_vm_scan_since: datetime.datetime | None = None,
        no_compliance_scan_since: datetime.datetime | None = None,
        vm_scan_since: datetime.datetime | None = None,
        compliance_scan_since: datetime.datetime | None = None,
        vm_processed_before: datetime.datetime | None = None,
        vm_processed_after: datetime.datetime | None = None,
        vm_scan_date_before: datetime.datetime | None = None,
        vm_scan_date_after: datetime.datetime | None = None,
        vm_auth_scan_date_before: datetime.datetime | None = None,
        vm_auth_scan_date_after: datetime.datetime | None = None,
        scap_scan_since: datetime.datetime | None = None,
        no_scap_scan_since: datetime.datetime | None = None,
        use_tags: bool | None = None,
        tag_set_by: Literal["id", "name"] | None = None,
        tag_include_selector: Literal["any", "all"] | None = None,
        tag_exclude_selector: Literal["any", "all"] | None = None,
        tag_set_include: list[str | int] | None = None,
        tag_set_exclude: list[str | int] | None = None,
        show_tags: bool | None = None,
        show_trurisk: bool | None = None,
        trurisk_min: int | None = None,
        trurisk_max: int | None = None,
        show_trurisk_factors: bool | None = None,
    ) -> tuple[list[host_list_output.Host], bool, int]:
        """Get a list of hosts from the VMDR API.  A value of None for the parameters will use their
            default values in the API.

        Args:
            ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.

        Returns:
            tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                host_list_vm_detection_output.HostList object, a boolean indicating whether the
                results were truncated, and the next id_min to use for the next call.
        """

        if ips is not None:
            if not isinstance(ips, list):
                ips = [ips]
            ipv4_ips: (
                list[str | ipaddress.IPv4Address | ipaddress.IPv6Address] | None
            ) = [
                ip
                for ip in ips
                if isinstance(ip, ipaddress.IPv4Address)
                or isinstance(ipaddress.ip_address(ip), ipaddress.IPv4Address)
            ]
            ipv6_ips: (
                list[str | ipaddress.IPv4Address | ipaddress.IPv6Address] | None
            ) = [
                ip
                for ip in ips
                if isinstance(ip, ipaddress.IPv6Address)
                or isinstance(ipaddress.ip_address(ip), ipaddress.IPv6Address)
            ]

            if not ipv4_ips:
                ipv4_ips = None
            if not ipv6_ips:
                ipv6_ips = None
        else:
            ipv4_ips = None
            ipv6_ips = None

        # Cannot pass both IPv4 and IPv6 addresses in the same call.
        if ipv4_ips and ipv6_ips:
            raise ValueError(
                "Cannot pass both IPv4 and IPv6 addresses in the same call."
            )

        params: dict[str, Any] = {
            "show_asset_id": show_asset_id,
            "details": details,
            "os_pattern": os_pattern,
            "truncation_limit": truncation_limit,
            "ips": ipv4_ips,
            "ipv6": ipv6_ips,
            "ag_ids": ag_ids,
            "ag_titles": ag_titles,
            "ids": ids,
            "id_min": id_min,
            "id_max": id_max,
            "network_ids": network_ids,
            "compliance_enabled": compliance_enabled,
            "no_vm_scan_since": no_vm_scan_since,
            "no_compliance_scan_since": no_compliance_scan_since,
            "vm_scan_since": vm_scan_since,
            "compliance_scan_since": compliance_scan_since,
            "vm_processed_before": vm_processed_before,
            "vm_processed_after": vm_processed_after,
            "vm_scan_date_before": vm_scan_date_before,
            "vm_scan_date_after": vm_scan_date_after,
            "vm_auth_scan_date_before": vm_auth_scan_date_before,
            "vm_auth_scan_date_after": vm_auth_scan_date_after,
            "scap_scan_since": scap_scan_since,
            "no_scap_scan_since": no_scap_scan_since,
            "use_tags": use_tags,
            "tag_set_by": tag_set_by,
            "tag_include_selector": tag_include_selector,
            "tag_exclude_selector": tag_exclude_selector,
            "tag_set_include": tag_set_include,
            "tag_set_exclude": tag_set_exclude,
            "show_tags": show_tags,
            "show_trurisk": show_trurisk,
            "trurisk_min": trurisk_min,
            "trurisk_max": trurisk_max,
            "show_trurisk_factors": show_trurisk_factors,
        }
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.host_list, params=params_cleaned).text
        match = re.search(
            r"<HOST_LIST_OUTPUT>.*?</HOST_LIST_OUTPUT>", raw_response, re.DOTALL
        )
        if match is None:
            raise ValueError("Cannot find HOST_LIST_OUTPUT in response.")
        host_list_output_str = match.group(0)
        host_list_output_obj = host_list_output.HostListOutput.from_xml(
            host_list_output_str
        )
        if host_list_output_obj.response.host_list is None:
            raise ValueError("Response has no host_list")

        host_list = host_list_output_obj.response.host_list
        warning = host_list_output_obj.response.warning
        if warning is None:
            return host_list, False, 0
        next_id_match = re.search(r"id_min=(\d+)", warning.url)
        if next_id_match is None:
            raise ValueError(
                "Unable to parse URL in warning message. No id_min found.\n"
                f"{warning.url}"
            )
        truncated = True
        next_id_min = int(next_id_match.group(1))

        return host_list, truncated, next_id_min

    def host_list_vm_detection(
        self,
        *,
        ids: int | list[int] | None = None,
        truncation_limit: int | None = None,
        id_min: int | None = None,
        qids: int | list[int] | None = None,
        show_qds: bool | None = None,
        qds_min: int | None = None,
        qds_max: int | None = None,
        arf_kernel_filter: int | None = None,
        show_igs: bool | None = None,
    ) -> tuple[list[host_list_vm_detection_output.Host], bool, int]:
        """Get a list of hosts with associated vulnerability detections from the VMDR API.  A
        value of None for the parameters will use their default values in the API.

            Args:
                ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.
                truncation_limit (int | None, optional): Maximum number of hosts to return. Defaults
                     to None.
                id_min (int | None, optional): Minimum host list ID to return. Defaults to None.

            Returns:
                tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                    host_list_vm_detection_output.HostList object, a boolean indicating whether the
                    results were truncated, and the next id_min to use for the next call.
        """

        params = {
            "ids": ids,
            "truncation_limit": truncation_limit,
            "id_min": id_min,
            "qids": qids,
            "show_qds": show_qds,
            "qds_min": qds_min,
            "qds_max": qds_max,
            "arf_kernel_filter": arf_kernel_filter,
            "show_igs": show_igs,
        }
        cleaned_params = qutils.clean_dict(params)
        cleaned_params["action"] = "list"

        raw_response = self.get(URLS.host_list_vm_detection, params=cleaned_params).text
        match = re.search(
            r"<HOST_LIST_VM_DETECTION_OUTPUT>.*?</HOST_LIST_VM_DETECTION_OUTPUT>",
            raw_response,
            re.DOTALL,
        )
        if match is None:
            raise ValueError("Cannot find HOST_LIST_VM_DETECTION_OUTPUT in response.")
        host_list_vm_detection_output_str = match.group(0)
        host_list_vm_detection_output_obj = (
            host_list_vm_detection_output.HostListVMDetectionOutput.from_xml(
                host_list_vm_detection_output_str
            )
        )
        if host_list_vm_detection_output_obj.response.host_list is None:
            raise ValueError("Response has no host_list")

        host_list = host_list_vm_detection_output_obj.response.host_list
        warning = host_list_vm_detection_output_obj.response.warning
        if warning is None:
            return host_list, False, 0
        next_id_match = re.search(r"id_min=(\d+)", warning.url)
        if next_id_match is None:
            raise ValueError(
                "Unable to parse URL in warning message. No id_min found.\n"
                f"{warning.url}"
            )
        truncated = True
        next_id_min = int(next_id_match.group(1))

        return host_list, truncated, next_id_min

    def knowledgebase(
        self,
        *,
        details: str | None = None,
        ids: int | list[int] | None = None,
        id_min: int | None = None,
        id_max: int | None = None,
    ) -> list[knowledgebase_output.Vuln]:
        """Get a list of vulnerabilities from the VMDR API.  A value of None for the parameters
        will use their default values in the API.

            Args:
                details (str | None, optional): Details to return. Defaults to None.
                ids (int | list[int] | None, optional): Vulnerability IDs to query. Defaults to
                  None.

            Returns:
                knowledge_base_vuln_list_output.VulnList: A VulnList object containing the list of
                    vulnerabilities.
        """
        params = {"details": details, "ids": ids, "id_min": id_min, "id_max": id_max}
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.knowledgebase, params=params_cleaned).text
        match = re.search(
            r"<KNOWLEDGE_BASE_VULN_LIST_OUTPUT>.*?</KNOWLEDGE_BASE_VULN_LIST_OUTPUT>",
            raw_response,
            re.DOTALL,
        )
        if match is None:
            raise ValueError("Cannot find KNOWLEDGE_BASE_VULN_LIST_OUTPUT in response.")
        knowledge_base_output_str = match.group(0)
        knowledge_base_output_obj = knowledgebase_output.KnowledgeBaseOutput.from_xml(
            knowledge_base_output_str
        )
        if knowledge_base_output_obj.response.vuln_list is None:
            raise ValueError("Response has no vuln_list")

        return knowledge_base_output_obj.response.vuln_list

    def launch_vm_scan(
        self,
        *,
        scan_title: str | None = None,
        iscanner_name: str | None = None,
        option_title: str | None = None,
        fqdn: str | list[str] | None = None,
    ) -> simple_return.SimpleReturn:
        params = {
            "scan_title": scan_title,
            "iscanner_name": iscanner_name,
            "option_title": option_title,
            "fqdn": fqdn,
        }
        params["action"] = "launch"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.post(
            URLS.launch_vm_scan, params=params_cleaned
        ).text.encode("utf-8")
        launch_vm_scan_output_obj = simple_return.SimpleReturn.from_xml(raw_response)

        if launch_vm_scan_output_obj.response.text != "New vm scan launched":
            raise ValueError(
                f"Failed to launch VM scan.  Response: {launch_vm_scan_output_obj.response.text}"
            )

        return launch_vm_scan_output_obj

    def map_report_list(
        self, *, last: bool | None = None, domain: str | None = None
    ) -> map_report_list.MapReportList:
        params = {"domain": domain}
        if last:
            params["last"] = "yes"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(
            URLS.map_report_list, params=params_cleaned
        ).text.encode("utf-8")
        map_report_list_output_obj = map_report_list.MapReportList.from_xml(
            raw_response
        )

        return map_report_list_output_obj

    def download_saved_map_report(self, *, ref: str | None) -> map_report.Map:
        params = {"ref": ref}
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(
            URLS.download_saved_map_report, params=params_cleaned
        ).text.encode("utf-8")
        map_report_output_obj = map_report.Map.from_xml(raw_response)

        return map_report_output_obj


class HostListORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List ORM Class.  Contains methods for loading hosts into an ORM database."""

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            load_func (Any): Function to call to get hosts.
            **kwargs (Any): Keyword arguments to pass to host_list.
        """

        def load_set(to_load: list[host_list_orm.Host]) -> None:
            """Load a single set of hosts into the ORM database.

            Args:
                to_load (list[host_list_orm.Host]): List of hosts to load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 10000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            hosts, truncated, next_id_min = self.host_list(**kwargs)
            to_load = [qutils.to_orm_object(host, host_list_orm.Host) for host in hosts]
            load_set(to_load)


class HostListVMDetectionORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List Detection ORM Class.  Contains methods for loading host
    detections into an ORM database.
    """

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List Detection ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_vm_detection_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            load_func (Any): Function to call to get hosts.
            **kwargs (Any): Keyword arguments to pass to host_list.
        """

        def load_set(to_load: list[host_list_vm_detection_orm.Host]) -> None:
            """Load a single set of hosts into the ORM database.

            Args:
                to_load (list[host_list_vm_detection_orm.Host]): List of hosts to load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 1000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            hosts, truncated, next_id_min = self.host_list_vm_detection(**kwargs)
            to_load = [
                qutils.to_orm_object(host, host_list_vm_detection_orm.Host)
                for host in hosts
            ]
            load_set(to_load)


class KnowledgebaseORM(VmdrAPI, QualysORMMixin):
    def __init__(self, echo: bool = False) -> None:
        VmdrAPI.__init__(self)
        self.orm_base = knowledgebase_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        vulns = self.knowledgebase(**kwargs)
        to_load = [qutils.to_orm_object(vuln, knowledgebase_orm.Vuln) for vuln in vulns]
        with orm.Session(self.engine) as session:
            for vuln in to_load:
                session.merge(vuln)
                
            session.commit()



# mypy: ignore-errors
# type: ignore

import inspect
import os
import sys
import unittest

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from qualyspy import asset_mgmt_tagging  # noqa: E402
from qualyspy.models.asset_mgmt_tagging import asset_request  # noqa: E402


class TestTags(unittest.TestCase):
    def test_add_update_search_tag(self):
        api = asset_mgmt_tagging.AssetMgmtTaggingAPI()
        add_resp = api.create_tag(
            name="Parent Tag",
            color="#FFFFFF",
            children=["Child 1", "Child 2", "Child 3"],
        )

        self.assertEqual(add_resp.response_code, "SUCCESS")

        parent_tag_id = add_resp.data[0].tag.id

        update_resp = api.update_tag(
            tag_id=parent_tag_id,
            add_children=["Child 4"],
        )

        self.assertEqual(update_resp.response_code, "SUCCESS")

        search_resp = api.search_tags(name="Parent Tag")

        self.assertEqual(search_resp.response_code, "SUCCESS")

        delete_resp = api.delete_tag(parent_tag_id)

        self.assertEqual(delete_resp.response_code, "SUCCESS")

    def test_asset(self):
        api = asset_mgmt_tagging.AssetMgmtTaggingAPI()
        get_resp = api.get_asset_info(asset_id=14355608)

        self.assertEqual(get_resp.response_code, "SUCCESS")

        update_resp = api.update_asset(
            asset_id=14355608,
            name="taijitu.private",
            add_tags=[12158745],
        )

        self.assertEqual(update_resp.response_code, "SUCCESS")

    def test_search_assets(self):
        api = asset_mgmt_tagging.AssetMgmtTaggingAPI()
        criteria = asset_request.Criteria(
            field="tagName", operator="EQUALS", value="Test Search Assets"
        )

        search_resp = api.search_assets(criteria=[criteria])
        
        self.assertEqual(search_resp.count, 1)



from . import URLS
from .base import QualysAPIBase
from .models.asset_mgmt_tagging import (
    asset_output,
    asset_request,
    tag_output,
    tag_request,
)
from .models.asset_mgmt_tagging.asset_request import Criteria as AssetSearchCriteria


class AssetMgmtTaggingAPI(QualysAPIBase):
    def create_tag(
        self,
        name: str,
        rule_type: tag_request.tag_rule_types | None = None,
        rule_text: str | None = None,
        criticality_score: int | None = None,
        color: str | None = None,
        children: list[str] = [],
    ) -> tag_output.ServiceResponse:
        request_data = tag_request.create_add_tag_request(
            name=name,
            rule_type=rule_type,
            rule_text=rule_text,
            criticality_score=criticality_score,
            color=color,
            children=children,
        )
        request_data_xml = request_data.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8", xml_declaration=True
        )

        resp = self.post(
            URLS.create_tag, content=request_data_xml, content_type="application/xml"
        )
        ret = tag_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def update_tag(
        self,
        tag_id: int,
        name: str | None = None,
        criticality_score: int | None = None,
        rule_type: tag_request.tag_rule_types | None = None,
        rule_text: str | None = None,
        color: str | None = None,
        add_children: list[str] = [],
        remove_children: list[int] = [],
    ) -> tag_output.ServiceResponse:
        if add_children and remove_children:
            raise ValueError(
                "Cannot add and remove children at the same time. Please use separate calls."
            )
        request_data = tag_request.create_update_tag_request(
            name=name,
            rule_type=rule_type,
            rule_text=rule_text,
            criticality_score=criticality_score,
            color=color,
            add_children=add_children,
            remove_children=remove_children,
        )
        request_data_xml = request_data.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8", xml_declaration=True
        )

        resp = self.post(
            URLS.update_tag + f"/{tag_id}",
            content=request_data_xml,
            content_type="application/xml",
        )
        ret = tag_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def search_tags(
        self,
        id: int | None = None,
        name: str | None = None,
        parent: int | None = None,
        rule_type: tag_request.tag_rule_types | None = None,
        provider: tag_request.tag_provider_types | None = None,
        color: str | None = None,
    ) -> tag_output.ServiceResponse:
        request_data = tag_request.create_search_tags_request(
            id=id,
            name=name,
            parent=parent,
            rule_type=rule_type,
            provider=provider,
            color=color,
        )
        request_data_xml = request_data.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8", xml_declaration=True
        )

        resp = self.post(
            URLS.search_tags,
            content=request_data_xml,
            content_type="application/xml",
        )
        ret = tag_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def delete_tag(self, tag_id: int) -> tag_output.ServiceResponse:
        resp = self.post(URLS.delete_tag + f"/{tag_id}")
        ret = tag_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def get_asset_info(self, asset_id: int) -> asset_output.ServiceResponse:
        resp = self.get(URLS.get_asset_info + f"/{asset_id}", accept="application/json")
        ret = asset_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def update_asset(
        self,
        asset_id: int | None,
        criteria: list[AssetSearchCriteria] | None = None,
        name: str | None = None,
        add_tags: list[int] = [],
        remove_tags: list[int] = [],
        start_from_offset: int | None = None,
        start_from_id: int | None = None,
        limit_results: int | None = None,
    ) -> asset_output.ServiceResponse:
        if add_tags and remove_tags:
            raise ValueError(
                "Cannot add and remove tags at the same time. Please use separate calls."
            )
        request_data = asset_request.create_asset_request(
            criteria=criteria,
            name=name,
            add_tags=add_tags,
            remove_tags=remove_tags,
            start_from_offset=start_from_offset,
            start_from_id=start_from_id,
            limit_results=limit_results,
        )
        request_data_xml = request_data.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8", xml_declaration=True
        )

        resp = self.post(
            URLS.update_asset + f"/{asset_id}",
            content=request_data_xml,
            content_type="application/xml",
        )
        ret = asset_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response

    def search_assets(
        self,
        criteria: list[AssetSearchCriteria],
        start_from_offset: int | None = None,
        start_from_id: int | None = None,
        limit_results: int | None = None,
    ) -> asset_output.ServiceResponse:
        request_data = asset_request.create_asset_request(
            criteria=criteria,
            start_from_offset=start_from_offset,
            start_from_id=start_from_id,
            limit_results=limit_results,
        )
        request_data_xml = request_data.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8", xml_declaration=True
        )

        resp = self.post(
            URLS.search_assets,
            content=request_data_xml,
            content_type="application/xml",
        )
        ret = asset_output.Wrapper.model_validate_json(resp.text)

        return ret.service_response



"""Base classes for intereacting with the Qualys API and the database where information is being
stored.  Primarily used internally by the QualysPy library, but can be used to call APIs which the
library does not yet support.

Typical usage example:
api = QualysAPIBase()
api.get("/msp/about.php")
"""

# For SQLAlchemy:
# mypy: allow-untyped-calls

import datetime
import urllib.parse
from abc import ABC, abstractmethod
from typing import Any

import httpx
import sqlalchemy as sa
import sqlalchemy.orm as orm
from decouple import config  # type: ignore

from . import URLS, exceptions

_USE_API_SERVER = ["msp", "api", "qps"]
_USE_API_GATEWAY = ["rest", "certview"]

_TIMEOUT = httpx.Timeout(120.0, read=300.0)


class QualysAPIBase:
    """Base class for interacting with the Qualys API.  This class is not intended to be used
    directly, but rather to be subclassed by other classes which implement specific Qualys API
    calls.

    Attributes:
        api_root (str): Root URL of the Qualys API.
        username (str): Username to use when authenticating to the Qualys API.
        password (str): Password to use when authenticating to the Qualys API.
        x_requested_with (str): Value to send in the X-Requested-With header.
        ratelimit_limit (int): Maximum number of requests allowed in the current window. Updated
            after every API call.
        ratelimit_window_sec (int): Length of the current window in seconds. Updated after every
            API call.
        ratelimit_remaining (int): Number of requests remaining in the current window. Updated
            after every API call.
        ratelimit_towait_sec (int): Number of seconds to wait before making another API call.
            Updated after every API call.
        concurrency_limit_limit (int): Maximum number of concurrent requests allowed. Updated
            after every API call.
    """

    def __init__(
        self,
        x_requested_with: str = "QualysPy Python Library",
    ) -> None:
        """Initializes an instance of the QualysAPIBase class.

        Args:
            config_file (str, optional): Path to the config file.  Defaults to
                ~/.qualyspy.
            x_requested_with (str, optional): Value to send in the X-Requested-With header.

        Raises:
            exceptions.ConfigError: Raised if the config file is missing a required key.
        """

        # Read config file
        self.api_server = str(config("QUALYS_API_SERVER"))
        self.api_gateway = str(config("QUALYS_API_GATEWAY"))
        self.username = str(config("QUALYS_USERNAME"))
        self.password = str(config("QUALYS_PASSWORD"))

        self.jwt: str | None = None

        self.x_requested_with = x_requested_with
        self.orm_base = orm.DeclarativeBase()

        self.ratelimit_limit: int | None = None
        self.ratelimit_window_sec: int | None = None
        self.ratelimit_remaining: int | None = None
        self.ratelimit_towait_sec: int | None = None
        self.concurrency_limit_limit: int | None = None
        self.concurrency_limit_running: int | None = None
        self.get(URLS.about)  # Set ratelimit and concurrency limit

    def _choose_url(self, url: str) -> str:
        """Choose the correct URL to use based on the URL.

        Args:
            url (str): URL to of endpoint being access.

        Returns:
            str: URL to use.
        """

        api_root = url.split("/")[1]

        if api_root in _USE_API_SERVER:
            return self.api_server
        elif api_root in _USE_API_GATEWAY:
            return self.api_gateway
        else:
            raise ValueError("No valid API root or gateway found.")

    def _get_jwt(self) -> None:
        """Get a new JWT from the Qualys API."""
        response = httpx.post(
            self.api_gateway + URLS.gateway_auth,
            data={
                "token": "true",
                "permissions": "true",  # Needed by CertView
                "username": self.username,
                "password": self.password,
            },
            headers={
                "X-Requested-With": self.x_requested_with,
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )
        try:
            response.raise_for_status()
        except httpx.HTTPError as e:
            raise exceptions.QualysAPIError(response.text) from e
        self.jwt = response.text

    def _update_limits(self, response: httpx.Response) -> None:
        """Update the rate limit and concurrency limit information.

        Args:
            response (requests.Response): Response object from the Qualys API.
        """

        def _get_updated_limit(headers: Any, name: str) -> int | None:
            """Get the updated limit or return None if the header is not present.

            Args:
                limit (str): Header to get the limit for.
            """

            try:
                return int(headers[name])
            except KeyError:
                return None

        self.ratelimit_limit = _get_updated_limit(response.headers, "X-RateLimit-Limit")
        self.ratelimit_window_sec = _get_updated_limit(
            response.headers, "X-RateLimit-Window"
        )
        self.ratelimit_towait_sec = _get_updated_limit(
            response.headers, "X-RateLimit-ToWait"
        )
        self.ratelimit_remaining = _get_updated_limit(
            response.headers, "X-RateLimit-Remaining"
        )
        self.concurrency_limit_limit = _get_updated_limit(
            response.headers, "X-ConcurrencyLimit"
        )
        self.concurrency_limit_running = _get_updated_limit(
            response.headers, "X-ConcurrencyRunning"
        )

    def get(
        self,
        url: str,
        params: dict[str, str] | None = None,
        accept: str = "application/xml",
    ) -> httpx.Response:
        """Send a GET request to the Qualys API.

        Args:
            url (str): URL to send the request to.
            params (dict[str, str], optional): Parameters to send with the request.  Defaults to
                None, which means the API call will use the default parameters.

        Returns:
            The text of the response.

        Raises:
            exceptions.QualysAPIError: Raised if the Qualys API returns a non-200 response.
        """
        root = self._choose_url(url)
        headers = {
            "X-Requested-With": self.x_requested_with,
        }
        if root == self.api_server:
            headers["Accept"] = accept
        elif root == self.api_gateway:
            if self.jwt is None:
                self._get_jwt()
            headers["Authorization"] = f"Bearer {self.jwt}"
        else:
            raise ValueError("No valid API root or gateway found.")
        try:
            response = httpx.get(
                root + url,
                params=params,
                auth=(self.username, self.password)
                if root == self.api_server
                else None,
                headers=headers,
                timeout=_TIMEOUT,
            )
        except httpx.ReadTimeout as e:
            raise exceptions.QualysAPIError(
                f"""
                                            Request for {root + url} timed out.
                                            params: {params},
                                            headers: {headers},
                                            timestamp: {datetime.datetime.now()}
                                            """
            ) from e
        try:
            response.raise_for_status()
        except httpx.HTTPError as e:
            raise exceptions.QualysAPIError(response.text) from e

        self._update_limits(response)
        return response

    def post(
        self,
        url: str,
        *,
        params: dict[str, str] | None = None,
        content: str | bytes | None = None,
        data: dict[str, str] | None = None,
        files: dict[str, Any] | None = None,
        content_type: str | None = "application/json",
        accept: str = "application/json",
    ) -> httpx.Response:
        """Send a POST request to the Qualys API.

        Args:
            url (str): URL to send the request to.
            data (dict[str, str], optional): Data to send with the request.  Defaults to None,
                which means the API call will use the default parameters.

        Returns:
            The text of the response.

        Raises:
            exceptions.QualysAPIError: Raised if the Qualys API returns a non-200 response.
        """
        root = self._choose_url(url)
        headers: dict[str, str] = {
            "X-Requested-With": self.x_requested_with,
        }
        if content_type is not None:
            headers["Content-Type"] = content_type
        if root == self.api_server:
            headers["Accept"] = accept
        elif root == self.api_gateway:
            if self.jwt is None:
                self._get_jwt()
            headers["Authorization"] = f"Bearer {self.jwt}"
        else:
            raise ValueError("No valid API root or gateway found.")
        try:
            response = httpx.post(
                root + url,
                params=params,
                content=content,
                data=data,
                files=files,
                auth=(self.username, self.password)
                if root == self.api_server
                else None,
                headers=headers,
                timeout=_TIMEOUT,
            )
        except httpx.ReadTimeout as e:
            raise exceptions.QualysAPIError(
                f"""
                                            Request for {root + url} timed out.
                                            params: {params},
                                            data: {data!r},
                                            headers: {headers},
                                            timestamp: {datetime.datetime.now()}
                                            """
            ) from e
        try:
            response.raise_for_status()
        except httpx.HTTPError as e:
            raise exceptions.QualysAPIError(response.text) from e

        self._update_limits(response)
        return response


class QualysORMMixin(ABC):
    """Mixin class for Qualys API classes that use SQLAlchemy ORM.

    Attributes:
        api (QualysAPIBase): Instance of the QualysAPIBase class.
        orm_base (sqlalchemy.ext.declarative.api.DeclarativeMeta): Base class for all ORM classes.
        db_host (str): Hostname of the PostgreSQL database.
        db_name (str): Name of the PostgreSQL database.
        db_username (str): Username to use to connect to the PostgreSQL database.
        db_password (str): Password to use to connect to the PostgreSQL database.
        e_url (str): SQLAlchemy engine URL.
        engine (sqlalchemy.engine.base.Engine): SQLAlchemy engine.
        echo (bool): Whether or not to echo SQL statements to stdout. Defaults to False.  If changed
            after the engine is created, the engine will automatically update with the new value.
    """

    def __init__(self, api: QualysAPIBase, *, echo: bool = False) -> None:
        """Initializes an instance of the QualysORMMixin class.

        Args:
            api (QualysAPIBase): Instance of the QualysAPIBase class.
            echo (bool, optional): Whether or not to echo SQL statements to stdout.  Defaults to
                False.

        Raises:
            exceptions.ConfigError: Raised if the config file is missing a required key.
        """
        self.api = api
        self.orm_base = api.orm_base
        try:
            self.db_host = str(config("PG_HOST"))
            self.db_name = str(config("PG_DB"))
            self.db_username = str(config("PG_USERNAME"))
            self.db_password = urllib.parse.quote(str(config("PG_PASSWORD")))
        except KeyError as e:
            raise exceptions.ConfigError(f"Config file missing key: {e}")
        self.e_url = "postgresql+psycopg:"
        self.e_url += (
            f"//{self.db_username}:{self.db_password}@{self.db_host}/{self.db_name}"
        )
        self.engine = sa.create_engine(self.e_url, echo=echo)

        self.echo = echo

    def init_db(
        self,
    ) -> None:
        """Initialize the database.  Creates the schema and tables if they don't already exist."""
        with self.engine.connect() as conn:
            if self.orm_base.metadata.schema is None:
                raise exceptions.ConfigError("Schema not set in ORM base.")
            conn.execute(
                sa.schema.CreateSchema(
                    self.orm_base.metadata.schema, if_not_exists=True
                )
            )
            conn.commit()

        self.orm_base.metadata.create_all(self.engine)

    @abstractmethod
    def load(self, **kwargs: Any) -> None:
        """Load data into the database."""
        ...

    def drop(self) -> None:
        """Drop the database."""
        with self.engine.connect() as conn:
            if self.orm_base.metadata.schema is None:
                raise exceptions.ConfigError("Schema not set in ORM base.")
            conn.execute(
                sa.schema.DropSchema(
                    self.orm_base.metadata.schema, cascade=True, if_exists=True
                )
            )
            conn.commit()

    def query(self, stmt: Any, *, echo: bool = False) -> Any:
        """Execute a query against the database.

        Args:
            stmt (Any): SQLAlchemy statement to execute.
            echo (bool, optional): Whether or not to echo SQL statements to stdout.  Defaults to
                False.

        Returns:
            list[_C]: List of objects returned by the query.
        """

        with orm.Session(self.engine) as session:
            results = session.execute(stmt)
            return results.all()

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Set an attribute of the QualysORMMixin class.  If the attribute is "echo", the engine
        attribute is updated with the new value.

        Args:
            __name (str): Name of the attribute to set.
            __value (Any): Value to set the attribute to.
        """
        super().__setattr__(__name, __value)
        if __name == "echo":
            self.engine = sa.create_engine(self.e_url, echo=__value)



"""Shareable SQLAlchemy data types"""

import ipaddress

import sqlalchemy as sa
import sqlalchemy.dialects.postgresql


class IPv4AddressType(sa.types.TypeDecorator[sa.types.String]):
    """SQLAlchemy type for IPv4Address"""

    impl = sqlalchemy.dialects.postgresql.INET

    def process_bind_param(self, value, dialect):  # type: ignore
        if value is not None:
            if isinstance(value, list):
                return value[0].exploded
            else:
                return value.exploded
        else:
            return None

    def process_result_value(self, value, dialect):  # type: ignore
        if value is not None:
            return ipaddress.IPv4Address(value)
        else:
            return None


class IPv6AddressType(sa.types.TypeDecorator[sa.types.String]):
    """SQLAlchemy type for IPv6Address"""

    impl = sqlalchemy.dialects.postgresql.INET

    def process_bind_param(self, value, dialect):  # type: ignore
        if value is not None:
            if isinstance(value, list):
                return value[0].exploded
            else:
                return value.exploded
        else:
            return None

    def process_result_value(self, value, dialect):  # type: ignore
        if value is not None:
            return ipaddress.IPv6Address(value)
        else:
            return None


class IPAddressGenericType(sa.types.TypeDecorator[sa.types.String]):
    impl = sqlalchemy.dialects.postgresql.INET

    # Clears the following warning:
    # SAWarning: TypeDecorator IPAddressGenericType() will not produce a cache key because the
    # ``cache_ok`` attribute is not set to True.  This can have significant performance
    # implications including some performance degradations in comparison to prior SQLAlchemy
    # versions.  Set this attribute to True if this type object's state is safe to use in a
    # cache key, or False to disable this warning. (Background on this warning at:
    # https://sqlalche.me/e/20/cprf)
    cache_ok = True

    def process_bind_param(self, value, dialect):  # type: ignore
        if value is not None:
            if isinstance(value, list):
                return value[0].exploded
            else:
                return value.exploded
        else:
            return None

    def process_result_value(self, value, dialect):  # type: ignore
        if value is not None:
            return ipaddress.ip_address(value)
        else:
            return None



from . import URLS
from .base import QualysAPIBase


class CertViewAPI(QualysAPIBase):
    """Qualys CertView API Class.  Contains methods for interacting with the CertView API."""

    def add_bulk_external_sites(self, *, sites: list[str]) -> None:
        """Add a list of external sites to CertView.

        Args:
            sites (list[str]): A list of external sites to add to CertView.
        """

        # A single CSV file can contain up to 1000 records, so add in batches of 1000
        for i in range(0, len(sites), 1000):
            sites_batch = sites[i : i + 1000]
            csv_string = "Sites\n" + "\n".join(sites_batch) + "\n"
            files = {"file": ("sites.csv", csv_string, "text/csv")}

            self.post(
                URLS.add_bulk_external_sites,
                params={"action": "SAVE_AND_LAUNCH"},
                files=files,
                # The content type is set automatically by requests based on the file type
                content_type=None,
                accept="application/json",
            )



from typing import Any

import sqlalchemy.orm as orm

from . import URLS, qutils
from .base import QualysAPIBase, QualysORMMixin
from .exceptions import QualysAPIError
from .models.gav import asset_details_orm, asset_details_output


class GavAPI(QualysAPIBase):
    """Qualys VMDR API Class.  Contains methods for interacting with the VMDR API."""

    def _convert_ipaddress(self, ips: str | None) -> list[str] | None:
        """Converts a string of IPs into a list of IP strings, for easier validation in
        Pydantic.
        """
        if ips is None:
            return None
        elif "," in ips:
            return ips.split(", ")
        return [ips]

    def asset_details(self, *, asset_id: int) -> asset_details_output.AssetItem | None:
        params = {"assetId": asset_id}
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.asset_details, params=params_cleaned)
        if raw_response.status_code == 204:
            return None
        response_json = raw_response.json()
        if (
            response_json["assetListData"]["asset"][0]["networkInterfaceListData"]
            is not None
        ):
            for interface in response_json["assetListData"]["asset"][0][
                "networkInterfaceListData"
            ]["networkInterface"]:
                interface["addressIpV4"] = self._convert_ipaddress(
                    interface["addressIpV4"]
                )
                interface["addressIpV6"] = self._convert_ipaddress(
                    interface["addressIpV6"]
                )
        response = asset_details_output.AssetDetailsOutput(**response_json)
        return response.asset_list_data.asset[0]

    def all_asset_details(
        self,
        *,
        last_seen_asset_id: int | None = None,
        page_size: int | None = None,
    ) -> tuple[list[asset_details_output.AssetItem], bool, int | None]:
        params = {
            "lastSeenAssetId": last_seen_asset_id,
            "pageSize": page_size,
        }
        params_cleaned = qutils.clean_dict(params)
        raw_response = self.post(URLS.all_asset_details, params=params_cleaned).json()

        for asset in raw_response["assetListData"]["asset"]:
            if asset["networkInterfaceListData"] is not None:
                for interface in asset["networkInterfaceListData"]["networkInterface"]:
                    interface["addressIpV4"] = self._convert_ipaddress(
                        interface["addressIpV4"]
                    )
                    interface["addressIpV6"] = self._convert_ipaddress(
                        interface["addressIpV6"]
                    )
            if (
                asset["cloudProvider"] is not None
                and asset["cloudProvider"]["oci"] is not None
            ):
                if asset["cloudProvider"]["oci"]["tags"] is None:
                    asset["cloudProvider"]["oci"]["tags"] = []

        response = asset_details_output.AssetDetailsOutput(**raw_response)

        if response.response_code != "SUCCESS":
            raise QualysAPIError(
                f"Qualys API returned an error: {response.response_code}"
            )
        has_more = bool(response.has_more)
        return response.asset_list_data.asset, has_more, response.last_seen_asset_id


class AllAssetDetailsORM(GavAPI, QualysORMMixin):
    def __init__(self, echo: bool = False) -> None:
        GavAPI.__init__(self)
        self.orm_base = asset_details_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        def _load_set(to_load: list[asset_details_orm.AssetItem]) -> None:
            with orm.Session(self.engine) as session:
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("page_size", 300)
        has_more = True

        # This value can be set to a specific asset ID to start from for debugging.
        last_seen_asset_id = None

        while has_more:
            kwargs["last_seen_asset_id"] = last_seen_asset_id
            assets, has_more, last_seen_asset_id = self.all_asset_details(**kwargs)
            to_load = [
                qutils.to_orm_object(asset, asset_details_orm.AssetItem)
                for asset in assets
            ]

            # Useful for debugging
            # print(last_seen_asset_id)

            _load_set(to_load)



# mypy: ignore-errors
# type: ignore

import inspect
import os
import sys
import unittest

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from qualyspy import certview  # noqa: E402


class TestCertViewAPI(unittest.TestCase):
    def test_add_bulk_external_sites(self):
        api = certview.CertViewAPI()
        sites = [
            "www.uwaterloo.ca",
            "www.quest.uwaterloo.ca",
            "quest.uwaterloo.ca",
            "lib.uwaterloo.ca",
        ]
        api.add_bulk_external_sites(sites=sites)
